cimport numpy as np
import numpy as np
from libc.stdint cimport uint8_t, uint64_t
from libcpp cimport bool

from .errors import BeliefPropagationError

cdef class _BinaryBeliefPropagation:
    """Abstract class that all belief propagation classes should inherit from.
    Implements the `run` and `marginals` function.
    """
    def __dealloc__(self):
        del self.bp

    def run(
        self,
        double precision=.1,
        int max_iter=100,
        double log_bound=50.,
        bool taylor_approximation=False,
        int n_threads=1
    ):
        """Run belief propagation.

        Args:
            precision: Stop iteration once the desired precision has been reached.
                Precision is defined as the l2-norm of the previous iteration of
                messages vs this iteration of messages.
            max_iter: Maximum number of iterations.
            log_bound: Clip the log-messages within [-log_bound, log_bound] to
                prevent under/over-flow.
            taylor_approximation: Use Taylor approximation to compute messages.
                May cause slight loss of accuracy.
            n_threads: Number of threads to use.
        """
        self.bp.run(precision, max_iter, log_bound, taylor_approximation, n_threads)

    def marginals(self, double log_bound=50.):
        """Compute the marginal probability of being in state 1 of all nodes.

        Returns:
            A 1-dimensional Numpy array of the same length as the number of
            nodes. This is essentially the flattened multi-dimensional array.
        """
        cdef np.ndarray[np.double_t, ndim=1] marginals = np.empty(self.n_nodes)
        self.bp.marginals(&marginals[0], log_bound)
        return marginals


cdef class FastBinaryBeliefPropagation(_BinaryBeliefPropagation):
    """Belief propagation on an arbitrary graph with binary variables.

    Args:
        n_nodes: Number of nodes in the graph.
        n_neighbors: Number of neighbors for each node. Dtype must be `np.uint8_t`.
        neighbors: A 1-dimensional flattened array of neighbors of each node.
            Dtype must be `np.uint64_t`.
        potentials0: 1D Numpy array of node potentials corresponding to state 0.
            This should be the array produced by calling `.flatten()` on the
            actual N-dimensional array.
        potentials1: 1D Numpy array of node potentials corresponding to state 1.
            This should be the array produced by calling `.flatten()` on the
            actual N-dimensional array.
        p: Edge potential when neighboring states are the same.
        q: Edge potential when neighboring states are different.
    """
    def __cinit__(
        self,
        uint64_t n_nodes,
        np.ndarray[uint8_t, ndim=1, mode='c'] n_neighbors,
        np.ndarray[uint64_t, ndim=1, mode='c'] neighbors,
        np.ndarray[double, ndim=1, mode='c'] potentials0,
        np.ndarray[double, ndim=1, mode='c'] potentials1,
        double p,
        double q
    ):
        self.n_nodes = n_nodes
        if self.n_nodes != potentials0.shape[0] or self.n_nodes != potentials1.shape[0]:
            raise BeliefPropagationError(
                f'Exactly {self.n_nodes} potentials must be provided.'
            )

        if n_neighbors.shape[0] != self.n_nodes:
            raise BeliefPropagationError(
                f'`n_neighbors` must have `n_nodes` elements'
            )
        if n_neighbors.sum() != neighbors.shape[0]:
            raise BeliefPropagationError(
                f'`neighbors` must have number of elements equal to the sum of '
                '`n_neighbors`'
            )

        try:
            self.bp = new BinaryBeliefPropagation(
                n_nodes,
                &n_neighbors[0],
                &neighbors[0],
                &potentials0[0],
                &potentials1[0],
                p,
                q
            )
        except Exception as e:
            raise BeliefPropagationError(e)



cimport cython
cimport numpy as np
import numpy as np
from libc.stdint cimport int16_t, uint32_t
from libc.stdlib cimport malloc, free

from .binary_belief_propagation cimport _BinaryBeliefPropagation
from .errors import BeliefPropagationError
from .wrappers.binary_belief_propagation cimport BinaryBeliefPropagation


cdef class FastBinaryGridBeliefPropagation(_BinaryBeliefPropagation):
    """Belief propagation on a grid MRF with binary variables. Has support for
    any-dimensional MRF.

    Args:
        shape: Shape of the grid as a Numpy array. Note that the dtype of the
            array must be `np.uint32`.
        neighbor_offsets: 2D Numpy array where each row corresponds to a
            potential neighbor, and the values correspond to the offests of
            each dimension from a node to that neighbor. Dtype must be `np.int16_t`.
        potentials0: 1D Numpy array of node potentials corresponding to state 0.
            This should be the array produced by calling `.flatten()` on the
            actual N-dimensional array.
        potentials1: 1D Numpy array of node potentials corresponding to state 1.
            This should be the array produced by calling `.flatten()` on the
            actual N-dimensional array.
        p: Edge potential when neighboring states are the same.
        q: Edge potential when neighboring states are different.
    """
    cdef int n_dims
    cdef np.ndarray shape

    def __cinit__(
        self,
        np.ndarray[uint32_t, ndim=1, mode='c'] shape,
        np.ndarray[int16_t, ndim=2, mode='c'] neighbor_offsets,
        np.ndarray[double, ndim=1, mode='c'] potentials0,
        np.ndarray[double, ndim=1, mode='c'] potentials1,
        double p,
        double q
    ):
        self.n_dims = shape.shape[0]
        self.shape = shape
        self.n_nodes = np.prod(shape)
        if self.n_nodes != potentials0.shape[0] or self.n_nodes != potentials1.shape[0]:
            raise BeliefPropagationError(
                f'Exactly {self.n_nodes} potentials must be provided.'
            )

        if neighbor_offsets.shape[1] != self.n_dims:
            raise BeliefPropagationError(
                f'`neighbor_offsets` must have {self.n_dims} as its second dimension.'
            )

        # Convert neighbor mask to C array
        cdef int max_neighbors = neighbor_offsets.shape[0]
        cdef int16_t** neighbor_offsets_p = <int16_t**>malloc(max_neighbors * sizeof(int16_t*))
        if not neighbor_offsets_p:
            raise MemoryError('Failed to allocate memory.')

        try:
            for i in range(max_neighbors):
                neighbor_offsets_p[i] = &neighbor_offsets[i, 0]

            self.bp = new BinaryBeliefPropagation(
                shape.shape[0],
                &shape[0],
                max_neighbors,
                &neighbor_offsets_p[0],
                &potentials0[0],
                &potentials1[0],
                p,
                q
            )
        except Exception as e:
            raise BeliefPropagationError(e)
        finally:
            free(neighbor_offsets_p)

    def marginals(self):
        """Compute the marginal probability of being in state 1 of all nodes.

        Returns:
            A N-dimensional Numpy array of the same shape as the shape provided
            upon initialization of this object.
        """
        return super().marginals().reshape(self.shape)



from libc.stdint cimport int16_t, uint8_t, uint16_t, uint32_t, uint64_t
from libcpp cimport bool

cdef extern from '../src/binary_belief_propagation.hpp':
    cdef cppclass BinaryBeliefPropagation:
        BinaryBeliefPropagation(uint8_t, const uint32_t*, uint8_t, int16_t**, const double*, const double*, double, double) except +
        BinaryBeliefPropagation(uint64_t, const uint8_t*, const uint64_t*, const double*, const double*, double, double) except +
        void run(double, uint16_t, double, bool, uint64_t)
        void marginals(double*, double);
